from typing import Tuple, List

from unsserv import extreme, stable
from unsserv.common.services_abc import (
    IMembershipService,
    IClusteringService,
    IAggregationService,
    ISamplingService,
    IDisseminationService,
    ISearchingService,
)
from unsserv.common.structs import Node

__all__ = [
    "join_network",
    "get_aggregation_service",
    "get_clustering_service",
    "get_sampling_service",
    "get_dissemination_service",
    "get_searching_service",
]


async def join_network(
    host: Tuple[str, int], service_id: str, is_extreme=True, **config
) -> IMembershipService:
    node = Node(address_info=host)
    if "bootstrap_nodes" in config:
        config["bootstrap_nodes"] = _bootstrap_nodes_to_structs(
            config["bootstrap_nodes"]
        )
    membership: IMembershipService
    if is_extreme:
        membership = extreme.Newscast(node)
    else:
        membership = stable.HyParView(node)
    await membership.join(service_id, **config)
    return membership


async def get_clustering_service(
    membership: IMembershipService, service_id: str, is_extreme=True, **config
) -> IClusteringService:
    clustering: IClusteringService
    if is_extreme:
        clustering = extreme.TMan(membership)
    else:
        clustering = stable.XBot(membership)
    await clustering.join(service_id, **config)
    return clustering


async def get_aggregation_service(
    membership: IMembershipService, service_id: str, is_extreme=True, **config
) -> IAggregationService:
    aggregation: IAggregationService
    if is_extreme:
        aggregation = extreme.AntiEntropy(membership)
    else:
        aggregation = stable.AntiEntropy(membership)
    await aggregation.join(service_id, **config)
    return aggregation


async def get_sampling_service(
    membership: IMembershipService, service_id: str, is_extreme=True, **config
) -> ISamplingService:
    sampling: ISamplingService
    if is_extreme:
        sampling = extreme.MRWB(membership)
    else:
        sampling = stable.RWD(membership)
    await sampling.join(service_id, **config)
    return sampling


async def get_dissemination_service(
    membership: IMembershipService,
    service_id: str,
    is_extreme=True,
    many_to_many=True,
    **config
):
    dissemination: IDisseminationService
    if is_extreme:
        if many_to_many:
            dissemination = extreme.Lpbcast(membership)
        else:
            dissemination = extreme.Mon(membership)
    else:
        if many_to_many:
            dissemination = stable.Plumtree(membership)
        else:
            dissemination = stable.Brisa(membership)
    await dissemination.join(service_id, **config)
    return dissemination


async def get_searching_service(
    membership: IMembershipService, service_id: str, is_extreme=True, **config
):
    searching: ISearchingService
    if is_extreme:
        searching = extreme.KWalker(membership)
    else:
        searching = stable.ABloom(membership)
    await searching.join(service_id, **config)
    return searching


def _bootstrap_nodes_to_structs(boostrap_nodes: List[Tuple[str, int]]) -> List[Node]:
    new_boostrap_nodes = []
    for node in boostrap_nodes:
        new_boostrap_nodes.append(Node(address_info=node))
    return new_boostrap_nodes



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.stable.dissemination.one_to_many.typing import BroadcastLevel


FIELD_COMMAND = "one_to_many-command"
FIELD_LEVEL = "one_to_many-level"
FIELD_BROADCAST_DATA = "one_to_many-broadcast-data"


class BrisaCommand(IntEnum):
    SESSION = auto()
    PUSH = auto()
    IM_YOUR_CHILD = auto()
    BECOME_MY_PARENT = auto()


class BrisaTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == BrisaCommand.SESSION:
            level: BroadcastLevel = data[0]
            message_data = {
                FIELD_COMMAND: BrisaCommand.SESSION,
                FIELD_LEVEL: level,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.PUSH:
            broadcast_data: bytes = data[0]
            message_data = {
                FIELD_COMMAND: BrisaCommand.PUSH,
                FIELD_BROADCAST_DATA: broadcast_data,  # type:ignore
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.IM_YOUR_CHILD:
            message_data = {
                FIELD_COMMAND: BrisaCommand.IM_YOUR_CHILD,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == BrisaCommand.BECOME_MY_PARENT:
            message_data = {
                FIELD_COMMAND: BrisaCommand.BECOME_MY_PARENT,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == BrisaCommand.SESSION:
            level: BroadcastLevel = message.data[FIELD_LEVEL]
            return BrisaCommand.SESSION, [level]
        elif command == BrisaCommand.PUSH:
            data: bytes = message.data[FIELD_BROADCAST_DATA]
            return BrisaCommand.PUSH, [data]
        elif command == BrisaCommand.IM_YOUR_CHILD:
            return BrisaCommand.IM_YOUR_CHILD, []
        elif command == BrisaCommand.BECOME_MY_PARENT:
            return BrisaCommand.BECOME_MY_PARENT, []
        raise ValueError("Invalid Command")


class BrisaProtocol(AProtocol):
    def _get_new_transcoder(self):
        return BrisaTranscoder(self.my_node, self.service_id)

    async def session(self, destination: Node, level: BroadcastLevel) -> bool:
        message = self._transcoder.encode(BrisaCommand.SESSION, level)
        return await self._rpc.call_send_message(destination, message)

    async def push(self, destination: Node, data: bytes):
        message = self._transcoder.encode(BrisaCommand.PUSH, data)
        return await self._rpc.call_send_message(destination, message)

    async def im_your_child(self, destination: Node) -> bool:
        message = self._transcoder.encode(BrisaCommand.IM_YOUR_CHILD)
        return await self._rpc.call_send_message(destination, message)

    async def become_my_parent(self, destination: Node) -> BroadcastLevel:
        message = self._transcoder.encode(BrisaCommand.BECOME_MY_PARENT)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_session(self, handler: Handler):
        self._handlers[BrisaCommand.SESSION] = handler

    def set_handler_push(self, handler: Handler):
        self._handlers[BrisaCommand.PUSH] = handler

    def set_handler_im_your_child(self, handler: Handler):
        self._handlers[BrisaCommand.IM_YOUR_CHILD] = handler

    def set_handler_become_my_parent(self, handler: Handler):
        self._handlers[BrisaCommand.BECOME_MY_PARENT] = handler



from enum import IntEnum, auto
from typing import Tuple, Sequence

from unsserv.common.structs import Node
from unsserv.common.rpc.structs import Message
from unsserv.common.rpc.protocol import AProtocol, ITranscoder, Command, Data, Handler
from unsserv.extreme.dissemination.one_to_many.structs import Session, Broadcast

FIELD_COMMAND = "one_to_many-command"
FIELD_BROADCAST_ID = "one_to_many-broadcast-id"
FIELD_LEVEL = "one_to_many-level"
FIELD_BROADCAST_DATA = "one_to_many-broadcast-data"


class MonCommand(IntEnum):
    SESSION = auto()
    PUSH = auto()


class MonTranscoder(ITranscoder):
    def encode(self, command: Command, *data: Data) -> Message:
        if command == MonCommand.SESSION:
            session: Session = data[0]
            message_data = {
                FIELD_COMMAND: MonCommand.SESSION,
                FIELD_BROADCAST_ID: session.broadcast_id,
                FIELD_LEVEL: session.level,
            }
            return Message(self.my_node, self.service_id, message_data)
        elif command == MonCommand.PUSH:
            broadcast: Broadcast = data[0]
            message_data = {
                FIELD_COMMAND: MonCommand.PUSH,
                FIELD_BROADCAST_ID: broadcast.id,
                FIELD_BROADCAST_DATA: broadcast.data,
            }
            return Message(self.my_node, self.service_id, message_data)
        raise ValueError("Invalid Command")

    def decode(self, message: Message) -> Tuple[Command, Sequence[Data]]:
        command = message.data[FIELD_COMMAND]
        if command == MonCommand.SESSION:
            session = Session(
                broadcast_id=message.data[FIELD_BROADCAST_ID],
                level=message.data[FIELD_LEVEL],
            )
            return MonCommand.SESSION, [session]
        elif command == MonCommand.PUSH:
            broadcast = Broadcast(
                id=message.data[FIELD_BROADCAST_ID],
                data=message.data[FIELD_BROADCAST_DATA],
            )
            return MonCommand.PUSH, [broadcast]
        raise ValueError("Invalid Command")


class MonProtocol(AProtocol):
    def _get_new_transcoder(self):
        return MonTranscoder(self.my_node, self.service_id)

    async def session(self, destination: Node, session: Session) -> bool:
        message = self._transcoder.encode(MonCommand.SESSION, session)
        return await self._rpc.call_send_message(destination, message)

    async def push(self, destination: Node, broadcast: Broadcast):
        message = self._transcoder.encode(MonCommand.PUSH, broadcast)
        return await self._rpc.call_send_message(destination, message)

    def set_handler_session(self, handler: Handler):
        self._handlers[MonCommand.SESSION] = handler

    def set_handler_push(self, handler: Handler):
        self._handlers[MonCommand.PUSH] = handler



from unsserv.stable.membership.hyparview import HyParView
from unsserv.stable.clustering.x_bot import XBot
from unsserv.stable.aggregation.anti_entropy import AntiEntropy
from unsserv.stable.sampling.rwd import RWD
from unsserv.stable.dissemination.one_to_many.brisa import Brisa
from unsserv.stable.dissemination.many_to_many.plumtree import Plumtree
from unsserv.stable.searching.abloom import ABloom



from unsserv.extreme.membership.newscast import Newscast
from unsserv.extreme.clustering.t_man import TMan
from unsserv.extreme.aggregation.anti_entropy import AntiEntropy
from unsserv.extreme.sampling.mrwb import MRWB
from unsserv.extreme.dissemination.one_to_many.mon import Mon
from unsserv.extreme.dissemination.many_to_many.lpbcast import Lpbcast
from unsserv.extreme.searching.k_walker import KWalker
